# High-Tune: Post-Process

This program is used to post process HDR images saved in the CIE 1931 XYZ color
space, as generated by the [HTRDR](https://gitlab.com/meso-star/htcp/) program.

The input images are simple ASCII files where empty lines are ignored. The text
behind the `#` char is considered as comments. The first valid line of the file
must list 2 integers defining the width and the height of the image. Then, each
subsequent line must list 3 pairs of floating point data, each pair
representing the estimation and the uncertainty of the X, Y and Z component of
a pixel. The pixels are row major ordered, i.e. they are listed row by row.

## How to build

This program is compatible GNU/Linux 64-bits. It relies on the
[CMake](http://www.cmake.org) and the
[RCMake](https://gitlab.com/vaplv/rcmake/) packages to build. It also depends
on the [RSys](https://gitlab.com/vaplv/rsys/) library and on the
[OpenMP](http://www.openmp.org) 1.2 specification to parallelize its
computations.

First ensure that CMake is installed on your system. Then install the RCMake
package as well as the aforementioned prerequisites. Finally, generate the
project from the `cmake/CMakeLists.txt` file by appending to the
`CMAKE_PREFIX_PATH` variable the install directories of its dependencies. The
resulting project can be edited, built, tested and installed as any CMake
project. Refer to the [CMake](https://cmake.org/documentation) for further
informations on CMake.

## Release notes


### Version 0.4.1

- Fix MSVC build.
- Change install dir for html man.
- Stop building html man on Unix systems (still optionally available).

### Version 0.4

- Add the `gnuplot` parameter to the `-m` option. Once set, the result image is
  written as a gnuplot script rather than a PPM image. This script generates a
  PNG image with an embedded color ramp.
- Fix the tone map operator: the exposure term is now also applied to the white
  radiance to ensure that it effectively corresponds to the white color in the
  output image.

### Version 0.3

- Add the `-i` option that regroups all the options controlling the
  post-processing of the pixel color. The options `-e` and `-w` that defined
  the exposure and the white scale factor are thus removed.
- Add support of raw data visualisation through color palettes. The `-m` option
  enables the mapping of a pixel component to a color ramp. Both, the pixel
  component to visualise and the palette to use can be set through this option.
  The data range to visualise can also be fixed. When this option is defined,
  the `-v` option can be used to displays into the terminal the color ramp and
  its associated values.
- Remove the `-u` and `-T` options that are previously used to roughly
  visualise the per pixel uncertainties and the estimate of the per realisation
  time: the new `-m` option proposes a far more efficient alternative.

### Version 0.2.1

- Update how the image exposure is handled: the pixels are multiplied by
  the exposure priorly to their tone mapping. Previously, it was the tone
  mapped pixels that were multiplied by the exposure.
- Fix the man page that wrongly described the tone mapping operator.

### Version 0.2

- Fix the XYZ to sRGB conversion: the reference white was not correctly set.
- Fix the gamma correction of the linear sRGB color: there was an issue in the
  used formulae.

### Version 0.1

- Handle the update of the htrdr-image file format introduced by
  [htrdr](https://gitlab.com/meso-star/htrdr/) 0.1 that adds to each pixel the
  estimation of the per realisation path computation time.

## License

Copyright (C) 2018, 2019, 2029 [|Meso|Star>](https://www.meso-star.com)
<contact@meso-star.com>.  Copyright (C) 2018-2019 Centre National de la
Recherche (CNRS), Université Paul Sabatier
<contact-edstar@laplace.univ-tlse.fr>. htpp is free software released under the
GPL v3+ license: GNU GPL version 3 or later. You are welcome to redistribute it
under certain conditions; refer to the COPYING file for details.

