/* Copyright (C) 2016-2020 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s2d_device_c.h"
#include "s2d_geometry.h"
#include "s2d_line_segments.h"

#include <rsys/mem_allocator.h>

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static void
geometry_release(ref_T* ref)
{
  struct geometry* geom;
  struct s2d_device* dev;

  geom = CONTAINER_OF(ref, struct geometry, ref);
  dev = geom->dev;
  if(geom->lines) line_segments_ref_put(geom->lines);
  MEM_RM(dev->allocator, geom);
  S2D(device_ref_put(dev));
}

/*******************************************************************************
 * Non exported functions
 ******************************************************************************/
res_T
geometry_create(struct s2d_device* dev, struct geometry** out_geom)
{
  struct geometry* geom = NULL;
  res_T res = RES_OK;
  ASSERT(dev && out_geom);

  geom = (struct geometry*)MEM_CALLOC
    (dev->allocator, 1, sizeof(struct geometry));
  if(!geom) {
    res = RES_MEM_ERR;
    goto error;
  }
  ref_init(&geom->ref);
  S2D(device_ref_get(dev));
  geom->dev = dev;
  geom->name = S2D_INVALID_ID;
  geom->rtc = NULL;
  geom->rtc_id = RTC_INVALID_GEOMETRY_ID;
  geom->embree_outdated_mask = 0;
  geom->flip_contour = 0;
  geom->is_enabled = 1;
  geom->lines = NULL;

exit:
  *out_geom = geom;
  return res;
error:
  if(geom) {
    geometry_ref_put(geom);
    geom = NULL;
  }
  goto exit;
}

void
geometry_ref_get(struct geometry* geom)
{
  ASSERT(geom);
  ref_get(&geom->ref);
}

void
geometry_ref_put(struct geometry* geom)
{
  ASSERT(geom);
  ref_put(&geom->ref, geometry_release);
}

