/* Copyright (C) 2016-2020 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#ifndef S2D_SCENE_VIEW_C_H
#define S2D_SCENE_VIEW_C_H

#include "s2d_backend.h"
#include "s2d_scene_c.h"

#include <rsys/dynamic_array_uint.h>
#include <rsys/hash_table.h>
#include <rsys/list.h>
#include <rsys/ref_count.h>

/* Forward declarations */
struct s2d_scene_view;

/* Generate the htable_geom hash table */
#define HTABLE_NAME geom
#define HTABLE_DATA struct geometry*
#define HTABLE_KEY unsigned /* Id of the shape */
#include <rsys/hash_table.h>

/* Generate the darray_fltui dynamic array */
struct fltui { float flt; unsigned ui; };
#define DARRAY_NAME fltui
#define DARRAY_DATA struct fltui
#include <rsys/dynamic_array.h>

/* Generate the darray_geom_nprims array */
struct nprims_cdf { unsigned nprims, ishape; };
#define DARRAY_NAME nprims_cdf
#define DARRAY_DATA struct nprims_cdf
#include <rsys/dynamic_array.h>

struct s2d_scene_view {
  struct list_node node; /* Attachment point to the scene scene_views pool */

  struct htable_geom cached_geoms; /* Cached shape geometries */
  struct darray_fltui cdf; /* Unormalized CDF */
  struct darray_nprims_cdf nprims_cdf;

   /* Id of Shapes detached while the scnview is active */
  struct darray_uint detached_shapes;

  float lower[2], upper[2]; /* AABB of the scene */

  /* Callback attached to the sig_shape_detach signal of scn */
  scene_shape_cb_T on_shape_detach_cb;

  int mask; /* Combination of enum s3d_scene_view_flag */
  int rtc_scn_update; /* Define if Embree geometries were deleted/added */
  int rtc_commit; /* Define whether or not the Embree scene was committed */
  RTCScene rtc_scn; /* Embree scene */

  ref_T ref;
  struct s2d_scene* scn;
};

extern LOCAL_SYM void
rtc_hit_filter_wrapper
  (const struct RTCFilterFunctionNArguments* args);

extern LOCAL_SYM void
scene_view_destroy
  (struct s2d_scene_view* scnview);

static FINLINE struct geometry*
scene_view_geometry_from_embree_id
  (struct s2d_scene_view* scnview,
   const unsigned irtc)
{
  struct geometry* geom;
  RTCGeometry rtc_geom;
  ASSERT(scnview && irtc != RTC_INVALID_GEOMETRY_ID);
  rtc_geom = rtcGetGeometry(scnview->rtc_scn, irtc);
  ASSERT(rtc_geom);
  geom = rtcGetGeometryUserData(rtc_geom);
  ASSERT(geom);
  return geom;
}

#endif /* S2D_SCENE_VIEW_C_H */
