/* Copyright (C) 2016-2020 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s2d.h"
#include "test_s2d_utils.h"

#include <rsys/float2.h>

static int
filter_hit
  (const struct s2d_hit* hit,
   const float org[2],
   const float dir[2],
   void* ray_data,
   void* filter_data)
{
  (void)dir, (void)org;
  CHK(hit != NULL);
  CHK((intptr_t)filter_data == 0xDEADBEEF);
  if(!ray_data) return 0;
  return S2D_PRIMITIVE_EQ((struct s2d_primitive*)ray_data, &hit->prim);
}

int
main(int argc, char** argv)
{
  struct s2d_device* dev;
  struct s2d_shape* shape;
  struct s2d_scene* scn;
  struct s2d_scene_view* scnview;
  struct s2d_vertex_data vdata;
  struct s2d_hit hit;
  struct s2d_primitive prim, prim2;
  struct mem_allocator allocator;
  float org[2], dir[3], range[2];
  float N[3] = {0.f, 0.f, 0.f};
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHK(s2d_device_create(NULL, &allocator, 1, &dev) == RES_OK);
  CHK(s2d_shape_create_line_segments(dev, &shape) == RES_OK);
  CHK(s2d_scene_create(dev, &scn) == RES_OK);

  vdata.type = S2D_FLOAT2;
  vdata.usage = S2D_POSITION;
  vdata.get = line_segments_get_position;
  CHK(s2d_line_segments_setup_indexed_vertices
    (shape, square_nsegs, line_segments_get_ids, square_nverts, &vdata, 1,
     (void*)&square_desc) == RES_OK);

  #define SET_FILTER_FUNC s2d_line_segments_set_hit_filter_function
  CHK(SET_FILTER_FUNC(NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(SET_FILTER_FUNC(shape, NULL, NULL) == RES_OK);
  CHK(SET_FILTER_FUNC(NULL, filter_hit, NULL) == RES_BAD_ARG);
  CHK(SET_FILTER_FUNC(shape, filter_hit, NULL) == RES_OK);
  CHK(SET_FILTER_FUNC(shape, filter_hit, (void*)0xDEADBEEF) == RES_OK);
  #undef SET_FILTER_FUNC

  CHK(s2d_scene_attach_shape(scn, shape) == RES_OK);

  f2(org, 0.f, 0.f);
  f2(dir, 0.f, -1.f);
  f2(range, 0, FLT_MAX);

  CHK(s2d_scene_view_create(scn, S2D_TRACE, &scnview) == RES_OK);

  #define RT s2d_scene_view_trace_ray
  CHK(RT(NULL, NULL, NULL, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(scnview, NULL, NULL, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(NULL, org, NULL, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(scnview, org, NULL, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(NULL, NULL, dir, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(scnview, NULL, dir, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(NULL, org, dir, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(scnview, org, dir, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(NULL, NULL, NULL, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(scnview, NULL, NULL, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(NULL, org, NULL, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(scnview, org, NULL, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(NULL, NULL, dir, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(scnview, NULL, dir, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(NULL, org, dir, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(scnview, org, dir, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT(NULL, NULL, NULL, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(scnview, NULL, NULL, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(NULL, org, NULL, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(scnview, org, NULL, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(NULL, NULL, dir, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(scnview, NULL, dir, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(NULL, org, dir, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(scnview, org, dir, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(NULL, NULL, NULL, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(scnview, NULL, NULL, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(NULL, org, NULL, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(scnview, org, NULL, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(NULL, NULL, dir, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(scnview, NULL, dir, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT(NULL, org, dir, range, NULL, &hit) == RES_BAD_ARG);

  f2(dir, 0.f, -1.f);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(hit.prim.prim_id == 0);
  f2_normalize(N, hit.normal);
  CHK(f2_eq_eps(N, f2(dir, 0.f, 1.f), 1.e-6f) == 1);
  CHK(eq_epsf(hit.u, 0.5f, 1.e-6f) == 1);
  CHK(eq_epsf(hit.distance, 1.0f, 1.e-6f) == 1);

  prim = hit.prim;

  f2(dir, 0.f, -1.f);
  range[1] = 0.5f;
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 1);
  range[1] = FLT_MAX;

  f2(dir, 0.f, -1.f);
  CHK(RT(scnview, org, dir, range, &prim, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 1);

  f2(dir, -1.f, 0.f);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(hit.prim.prim_id == 1);
  f2_normalize(N, hit.normal);
  CHK(f2_eq_eps(N, f2(dir, 1.f, 0.f), 1.e-6f) == 1);
  CHK(eq_epsf(hit.u, 0.5f, 1.e-6f) == 1);
  CHK(eq_epsf(hit.distance, 1.0f, 1.e-6f) == 1);

  f2(dir, 0.f, 1.f);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(hit.prim.prim_id == 2);
  f2_normalize(N, hit.normal);
  CHK(f2_eq_eps(N, f2(dir, 0.f, -1.f), 1.e-6f) == 1);
  CHK(eq_epsf(hit.u, 0.5f, 1.e-6f) == 1);
  CHK(eq_epsf(hit.distance, 1.0f, 1.e-6f) == 1);

  prim2 = hit.prim;

  f2(org, 0.f, 2.f);
  f2(dir, 0.f, -1.f);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(S2D_PRIMITIVE_EQ(&hit.prim, &prim2) == 1);

  CHK(RT(scnview, org, dir, range, &prim2, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(S2D_PRIMITIVE_EQ(&hit.prim, &prim) == 1);

  f2_splat(org, 0.f);
  f2(dir, 1.f, 0.f);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(hit.prim.prim_id == 3);
  f2_normalize(N, hit.normal);
  CHK(f2_eq_eps(N, f2(dir, -1.f, 0.f), 1.e-6f) == 1);
  CHK(eq_epsf(hit.u, 0.5f, 1.e-6f) == 1);
  CHK(eq_epsf(hit.distance, 1.0f, 1.e-6f) == 1);

  f2(range, 1.f, -1.f);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 1);

  f2(dir, 1.f, 1.f);
  f2(range, 0.f, FLT_MAX);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_BAD_ARG);

  CHK(s2d_scene_view_ref_put(scnview) == RES_OK);

  CHK(s2d_scene_view_create(scn, S2D_TRACE, &scnview) == RES_OK);
  f2_sub(dir, f2(dir, 0.75f, -1.f), org);
  f2_normalize(dir, dir);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(eq_epsf(hit.u, 0.125f, 1.e-6f) == 1);
  CHK(eq_epsf(hit.distance, 1.25, 1.E-6f) == 1);

  f2_sub(dir, f2(dir, -1.f, 0.25f), org);
  f2_normalize(dir, dir);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(eq_epsf(hit.u, 0.625, 1.e-6f) == 1);
  CHK(eq_epsf(hit.distance, (float)sqrt(1.0625f), 1e-6f) == 1);

  f2(org, -1.25f, 0.f);
  f2(dir, -1, 0.f);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 1);

  f2(dir, 1, 0.f);
  CHK(RT(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  CHK(hit.prim.prim_id == 1);
  f2_normalize(N, hit.normal);
  CHK(f2_eq_eps(N, f2(dir, 1.f, 0.f), 1.e-6f) == 1);
  CHK(eq_epsf(hit.u, 0.5f, 1.e-6f) == 1);
  CHK(eq_epsf(hit.distance, 0.25f, 1.e-6f) == 1);

  CHK(s2d_scene_view_ref_put(scnview) == RES_OK);

  CHK(s2d_device_ref_put(dev) == RES_OK);
  CHK(s2d_shape_ref_put(shape) == RES_OK);
  CHK(s2d_scene_ref_put(scn) == RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHK(mem_allocated_size() == 0);

  return 0;
}

