/* Copyright (C) 2016-2020 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#ifndef S2D_UTILS_H
#define S2D_UTILS_H

#include <rsys/mem_allocator.h>
#include <stdio.h>

struct line_segments_desc {
  const float* vertices;
  const unsigned* indices;
};

/*******************************************************************************
 * Geometries
 ******************************************************************************/
static const float square_verts[] = {
  1.f, -1.f,
  -1.f, -1.f,
  -1.f, 1.f,
  1.f, 1.f
};
const unsigned square_nverts = sizeof(square_verts)/sizeof(float[2]);

const unsigned square_ids[] = {
  0, 1, /* Bottom */
  1, 2, /* Left */
  2, 3, /* Top */
  3, 0  /* Right */
};
const unsigned square_nsegs = sizeof(square_ids)/sizeof(unsigned[2]);

static const struct line_segments_desc square_desc = { square_verts, square_ids };

static const float line_verts[] = { -1.f, 0.f, 1.f, 0.f };
const unsigned line_nverts = sizeof(line_verts)/sizeof(float[2]);
const unsigned line_ids[] = { 0, 1 };
const unsigned line_nsegs = sizeof(line_ids)/sizeof(unsigned[2]);
static const struct line_segments_desc line_desc = { line_verts, line_ids };

static INLINE void
line_segments_get_ids(const unsigned isegment, unsigned ids[2], void* data)
{
  const unsigned id = isegment * 2;
  const struct line_segments_desc* desc = data;
  CHK(desc != NULL);
  ids[0] = desc->indices[id + 0];
  ids[1] = desc->indices[id + 1];
}

static INLINE void
line_segments_get_position(const unsigned ivert, float position[2], void* data)
{
  const unsigned id = ivert * 2;
  const struct line_segments_desc* desc = data;
  CHK(desc != NULL);
  position[0] = desc->vertices[id + 0];
  position[1] = desc->vertices[id + 1];
}

/*******************************************************************************
 * Miscellaneous function
 ******************************************************************************/
static INLINE float
rand_canonic(void)
{
  int r;
  while((r = rand()) == RAND_MAX);
  return (float)r / (float)RAND_MAX;
}

static INLINE void
check_memory_allocator(struct mem_allocator* allocator)
{
  if(MEM_ALLOCATED_SIZE(allocator)) {
    char dump[512];
    MEM_DUMP(allocator, dump, sizeof(dump)/sizeof(char));
    fprintf(stderr, "%s\n", dump);
    FATAL("Memory leaks\n");
  }
}

#endif /* S2D_UTILS_H */

