/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "sstl.h"
#include "test_sstl_utils.h"

#include <rsys/logger.h>

static void
log_stream(const char* msg, void* ctx)
{
  ASSERT(msg);
  (void)msg, (void)ctx;
  printf("%s\n", msg);
}

int
main(int argc, char** argv)
{
  struct logger logger;
  struct mem_allocator allocator;
  struct sstl* sstl;
  (void)argc, (void)argv;

  CHK(sstl_create(NULL, NULL, 0, NULL) == RES_BAD_ARG);
  CHK(sstl_create(NULL, NULL, 0, &sstl) == RES_OK);

  CHK(sstl_ref_get(NULL) == RES_BAD_ARG);
  CHK(sstl_ref_get(sstl) == RES_OK);
  CHK(sstl_ref_put(NULL) == RES_BAD_ARG);
  CHK(sstl_ref_put(sstl) == RES_OK);
  CHK(sstl_ref_put(sstl) == RES_OK);

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHK(sstl_create(NULL, &allocator, 0, NULL) == RES_BAD_ARG);
  CHK(sstl_create(NULL, &allocator, 0, &sstl) == RES_OK);
  CHK(sstl_ref_put(sstl) == RES_OK);

  CHK(logger_init(&allocator, &logger) == RES_OK);
  logger_set_stream(&logger, LOG_OUTPUT, log_stream, NULL);
  logger_set_stream(&logger, LOG_ERROR, log_stream, NULL);
  logger_set_stream(&logger, LOG_WARNING, log_stream, NULL);

  CHK(sstl_create(&logger, NULL, 0, NULL) == RES_BAD_ARG);
  CHK(sstl_create(&logger, NULL, 0, &sstl) == RES_OK);
  CHK(sstl_ref_put(sstl) == RES_OK);
  CHK(sstl_create(&logger, &allocator, 0, NULL) == RES_BAD_ARG);
  CHK(sstl_create(&logger, &allocator, 0, &sstl) == RES_OK);
  CHK(sstl_ref_put(sstl) == RES_OK);

  CHK(sstl_create(&logger, &allocator, 1, NULL) == RES_BAD_ARG);
  CHK(sstl_create(&logger, &allocator, 1, &sstl) == RES_OK);
  CHK(sstl_ref_put(sstl) == RES_OK);

  logger_release(&logger);
  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHK(mem_allocated_size() == 0);
  return 0;
}

