/* Copyright (C) 2020 |Meso|Star> (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#ifndef GREEN_DESCRIPTION_H
#define GREEN_DESCRIPTION_H

#include <rsys/str.h>

/* Different types of descriptions */
enum description_type {
  DESC_MAT_SOLID,
  DESC_MAT_FLUID,
  DESC_BOUND_H_FOR_FLUID,
  DESC_BOUND_H_FOR_SOLID,
  DESC_BOUND_T_FOR_FLUID,
  DESC_BOUND_T_FOR_SOLID,
  DESC_BOUND_F_FOR_SOLID,
  DESC_SOLID_FLUID_CONNECT,
  DESCRIPTION_TYPE_COUNT__,
  DESC_OUTSIDE
};

#define DESC_IS_H(D) \
  ((D) == DESC_BOUND_H_FOR_SOLID || (D) == DESC_BOUND_H_FOR_FLUID)
#define DESC_IS_T(D) \
  ((D) == DESC_BOUND_T_FOR_SOLID || (D) == DESC_BOUND_T_FOR_FLUID)

struct mat_fluid {
  struct str name;
  unsigned fluid_id;
  double rho;
  double cp;
  double tinit;
  double imposed_temperature;
};

struct mat_solid {
  struct str name;
  unsigned solid_id;
  double lambda;
  double rho;
  double cp;
  double delta;
  double tinit;
  double imposed_temperature;
  double vpower;
};

struct h_boundary {
  struct str name;
  unsigned mat_id;
  double emissivity;
  double specular_fraction;
  double hc;
  double imposed_temperature;
};

struct t_boundary {
  struct str name;
  unsigned mat_id;
  double emissivity;
  double specular_fraction;
  double hc;
  double imposed_temperature;
};

struct f_boundary {
  struct str name;
  unsigned mat_id;
  double imposed_flux;
};

struct solid_fluid_connect {
  struct str name;
  double emissivity;
  double specular_fraction;
  double hc;
};

struct description {
  enum description_type type;
  union {
    struct mat_fluid fluid;
    struct mat_solid solid;
    struct t_boundary t_boundary;
    struct f_boundary f_boundary;
    struct h_boundary h_boundary;
    struct solid_fluid_connect sf_connect;
  } d;
};

static INLINE const struct str*
get_description_name
  (const struct description* desc)
{
  ASSERT(desc);
  switch (desc->type) {
  case DESC_MAT_SOLID:
    return &desc->d.solid.name;
  case DESC_MAT_FLUID:
    return &desc->d.fluid.name;
  case DESC_BOUND_T_FOR_SOLID:
  case DESC_BOUND_T_FOR_FLUID:
    return &desc->d.t_boundary.name;
  case DESC_BOUND_H_FOR_SOLID:
  case DESC_BOUND_H_FOR_FLUID:
    return &desc->d.h_boundary.name;
  case DESC_BOUND_F_FOR_SOLID:
    return &desc->d.f_boundary.name;
  case DESC_SOLID_FLUID_CONNECT:
    return &desc->d.sf_connect.name;
  default:
    FATAL("error:" STR(__FILE__) ":" STR(__LINE__)": Invalid type.\n");
  }
}

#endif /* GREEN_DESCRIPTION_H */
